/*
    SPDX-FileCopyrightText: 2015 Ivan Cukic <ivan.cukic(at)kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#ifndef SPECIALVALUES_H
#define SPECIALVALUES_H

#include <QString>

#define GLOBAL_ACTIVITY_TAG QStringLiteral(":global")
#define ANY_ACTIVITY_TAG QStringLiteral(":any")
#define CURRENT_ACTIVITY_TAG QStringLiteral(":current")

#define GLOBAL_AGENT_TAG QStringLiteral(":global")
#define ANY_AGENT_TAG QStringLiteral(":any")
#define CURRENT_AGENT_TAG QStringLiteral(":current")

#define ANY_TYPE_TAG QStringLiteral(":any")
#define FILES_TYPE_TAG QStringLiteral(":files")
#define DIRECTORIES_TYPE_TAG QStringLiteral(":directories")

#endif // SPECIALVALUES_H
